# -*- coding: utf-8 -*-
#
# This file is part of Chronoscope.
#
# SPDX-FileCopyrightText: 2024 Anatoliy Bilenko <anatoliy.bilenko@gmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

import matplotlib.pyplot as pt          # type: ignore
from yaml import safe_load
from enum import Enum
from abc import abstractmethod

class orientation(Enum):
    VERTICAL = 1
    HORIZONTAL = 2

class drawer:
    @abstractmethod
    def one(self, ev_begin: str, ev_end: str, tk_type: str): pass  # noqa

    def __init__(self, fig: str, figsize: tuple, unit_of_time: str):
        conv = {"ms": 1e6, "us": 1e3}
        self.fig = fig
        self.conv = conv[unit_of_time] if unit_of_time in conv else 1
        self.figsize = figsize
        self.unit_of_time = unit_of_time

    def draw(self, yspans: str, orient=orientation.HORIZONTAL):
        spans: list[dict[str, str]] = safe_load(yspans)
        spans_nr = len(spans)
        pt.figure(figsize=self.figsize)
        for i, s in enumerate(spans):
            match orient:
                case orientation.HORIZONTAL:
                    pt.subplot(1, spans_nr, i + 1)
                case orientation.VERTICAL:
                    pt.subplot(spans_nr, 1, i + 1)

            type = s["type"]
            span = s["span"]
            self.one(span[0], span[1], type)
        pt.savefig(fname=self.fig)



# -*- coding: utf-8 -*-
#
# This file is part of Chronoscope.
#
# SPDX-FileCopyrightText: 2024 Anatoliy Bilenko <anatoliy.bilenko@gmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

import chronoscope.db as db
import matplotlib.pyplot as pt          # type: ignore
import matplotlib.ticker as ticker      # type: ignore
import chronoscope.utils as utils
import chronoscope.drawer as drawer


class queue(drawer.drawer):
    def __init__(self, fig="queue.svg", figsize=(12, 4), unit_of_time="ms"):
        super().__init__(fig, figsize, unit_of_time)

    def one(self, ev_begin: str, ev_end: str, tk_type: str):
        queues = [s for s in db.queues(ev_begin, ev_end, tk_type)]
        times = [x[0] for x in queues]
        in_flight = []
        in_flight_c = 0
        for x in queues:
            in_flight_c += x[1] - x[2]
            in_flight.append(in_flight_c)

        pt.title(f"{tk_type}: {ev_begin} — {ev_end}")
        pt.xlabel(f"time({self.unit_of_time})")
        pt.ylabel("in-flight")
        pt.tight_layout()
        pt.gca().xaxis.set_major_formatter(ticker.FuncFormatter(
            lambda value, pos: utils.str_ns(value, compact=True)))
        pt.plot(times, in_flight, 'r--', linewidth=1)



# -*- coding: utf-8 -*-
#
# This file is part of Chronoscope.
#
# SPDX-FileCopyrightText: 2024 Anatoliy Bilenko <anatoliy.bilenko@gmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

from chronoscope import db
from chronoscope import parser
import chronoscope.hist as hist
import chronoscope.tree as tree
import chronoscope.chart as chart
import chronoscope.queue as queue
import chronoscope.drawer as drawer
import argparse as arg
import errno
import sys

__version__ = '0.0.3'
__author__ = 'Anatoliy Bilenko <anatoliy.bilenko@gmail.com>'
__license__ = 'LGPLv3'

db_options: dict[str, int | str] = {
    "cache_size": -128 << 20,
    "synchronous": "off",
    "journal_mode": "off",
}

def parse_args():
    parser = arg.ArgumentParser(prog=sys.argv[0],
                                formatter_class=arg.RawTextHelpFormatter,
                                description="""
    chronoscope: Plots drill-down chart based on user's traces

    Examples:
    chronoscope hist -S \
    "[{type: conn, span:[started,stopped]}, {type: gw, span:[inited,closed]}]"
    """)
    parser.add_argument("-d", "--db", type=str, default="chronoscope.db",
                        help="chronoscope database")
    parser.add_argument("-c", "--conf", type=str, default="chronoscope.yaml",
                        help="configuration, defines how to parse out users'\n"
                        "ticks, attrs and relations from the traces")
    parser.add_argument("-t", "--trace", type=str, help="User's traces")
    parser.add_argument("-D", "--depth", type=int, default=50,
                        help="limits output to given level of ticks")
    parser.add_argument("command", type=str,
                        choices=["create", "chart", "tree", "hist", "queue"],
                        help="create: build chronoscope database\n"
                        "plot: plots drill-down chart for the given tick")
    parser.add_argument("-f", "--fig_size", nargs=2, type=int, default=[16, 4])
    parser.add_argument("-k", "--tick_id", type=int,
                        help="tick identifier to plot")
    parser.add_argument("-S", "--spans", type=str, default="[]",
                        help="histogram spans")
    return parser.parse_args()


def main() -> int:
    try:
        args = parse_args()
        match args.command:
            case "create":
                db.open(args.db, db_options, create=True)
                db.load(parser.parser(args.conf), args.trace)
                db.mkidx()
                db.close()
            case "chart":
                db.open(args.db, db_options)
                chart.plot(args.tick_id, args.fig_size, args.depth)
            case "tree":
                db.open(args.db, db_options)
                tree.plot(args.tick_id, args.depth)
            case "hist":
                db.open(args.db, db_options)
                hist.hist().draw(args.spans)
            case "queue":
                db.open(args.db, db_options)
                queue.queue().draw(args.spans, drawer.orientation.VERTICAL)
    except KeyboardInterrupt:
        return errno.EINTR
    except FileExistsError as e:
        print(e, file=sys.stderr)
        return errno.EEXIST
    except FileNotFoundError as e:
        print(e, file=sys.stderr)
        return errno.EACCES

    return 0



# -*- coding: utf-8 -*-
#
# This file is part of Chronoscope.
#
# SPDX-FileCopyrightText: 2024 Anatoliy Bilenko <anatoliy.bilenko@gmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

import chronoscope.db as db
import scipy.stats as ss                # type: ignore
import matplotlib.pyplot as pt          # type: ignore
from statistics import stdev
import chronoscope.drawer as drawer


class hist(drawer.drawer):
    def __init__(self, fig="hist.svg", figsize=(12, 4), unit_of_time="ms"):
        super().__init__(fig, figsize, unit_of_time)

    def one(self, ev_begin: str, ev_end: str, tk_type: str):
        spans = [s[0] / self.conv for s in db.spans(ev_begin, ev_end, tk_type)]
        tot = len(spans)
        smax = round(max(spans), 2)
        smin = round(min(spans), 2)
        savg = round(sum(spans) / tot, 2)
        stde = round(stdev(spans), 2)
        # rlim = [f for f in spans if options is None or f < options]
        rlim = [f for f in spans]
        rlim_tot = len(rlim)
        ratio_p = round(rlim_tot / tot, 2)

        s1 = fr"$i = [{smin}, {smax}], \mu={savg},\ \sigma={stde}$"
        s2 = fr"$n = {tot}, nr = {rlim_tot}, nr/n = {ratio_p}$"
        pt.title(f"{tk_type}: {ev_begin} — {ev_end}\n{s1}\n{s2}")
        pt.xlabel(f"time({self.unit_of_time})")
        pt.ylabel("frequency")
        n, bins, patches = pt.hist(rlim, bins=50, facecolor='green', alpha=0.7)
        y = ss.norm.pdf(bins, savg, stde)
        pt.tight_layout()
        pt.plot(bins, y, 'r--', linewidth=1)
